#!/usr/local/svi/bin/tclsh

# FILE: doBatch.tcl
#
# PURPOSE:
# Process images in batch, taking directions from a control file
#

#   Author and copyright: 
#   T. Dijkhuis/H.v.d.Voort Scientific Volume Imaging Holding BV
#   Alexanderlaan 14,
#   1213 XS Hilversum,
#   The Netherlands,
#   email: info@svi.nl



# ---------------------------- Small utilities -----------------------------


proc isHuygens {} {
    global G_runningHuygens

    if [catch { huOpt report ""} ] {
        set G_runningHuygens 0
        return 0
    } else {
        set G_runningHuygens 1
        return 1
    }
}

proc checkHuScriptVersion {major minor} {
    global G_huscriptVersion env
    
    set myExec [lindex $G_huscriptVersion 0]
    if {![file exists  $myExec]} {
        puts "Error: executable $myExec is missing"
        return 0
    }
    
    if [catch { open "|$G_huscriptVersion" r+} pipe] {
        puts "Failed to launch $G_huscriptVersion"
        return 0
    }
    if [eof $pipe] {
        puts "Failed to communicate with $G_huscriptVersion (1)"
        return 0
    }
    
    # Now read and discard all lines written by huscript to stdout at startup:
    puts $pipe 	"puts initDone"
    flush $pipe
    set ready 0
    set cnt 0
    set maxCnt 20
    while {$ready == 0 && $cnt < $maxCnt} {  
        if [eof $pipe] {
            puts "Failed to communicate with $myExec; check by executing \
                  \n\"$G_huscriptVersion\" from a shell. Also check license."
            if {[string length $env(SVIHOME)] > 0} {
		puts "The SVIHOME environment variable is set\
		    to \"$env(SVIHOME)\". Check whether this is\ncorrect."
	    }
	    return 0
        }
        gets $pipe pipeOut
        if {[string compare $pipeOut "initDone"] == 0} {
            set ready 1
        }
        incr cnt 1
    }  
    if {$cnt >= $maxCnt} {
        puts "Too much output from $G_huscriptVersion, last line: $pipeOut"
        return 0
    }
    
    puts $pipe 	"puts \[huOpt version\]"
    flush $pipe
    gets $pipe result

    puts $pipe 	"puts \[huOpt product\]"
    flush $pipe
    gets $pipe product
    
    puts $pipe 	"puts \[huOpt license\]"
    flush $pipe
    gets $pipe license

    # Shut down huscript:
    puts $pipe "exit"
    flush $pipe
    catch {close $pipe}
    
    scan $result "%d.%d%s" actualMajor actualMinor dummy
    if {$major == $actualMajor && $minor <= $actualMinor
        && [string compare $product "HuScript"] == 0} {
        puts "Found $myExec ($product) version $actualMajor.$actualMinor: ok."
    } else {
        puts "Script doBatch.tcl requires Huygens Scripting version\
              $major.$minor but\n$myExec reports $product is version\
              $actualMajor.$actualMinor ($result).\nCannot continue."
        return 0
    }

    if {[string first "freeware" $license] >= 0} {
        puts "Insufficient license found for Huygens Scripting: freeware mode"
        return 0
    } else {
        puts "Found license options: $license"
    }
    
    return 1
}

# Print a message depending on the 'verbosity' level. 
# If the 'verbosity' level is equal or lower than the global level it is printed; else not.
# The default value of the global verbosity level is 0. It can be set with the -v [level]
# option on the command line.
# If this script is executed by a Huygens-Tcl interprter output is sent to the task 
# report window, else to stdout. 
#
proc myprint { msg {verbosity 0}} {
    global G_runningHuygens G_verbosity

    if {$G_verbosity >= $verbosity} { 
        if {$G_runningHuygens} {
            huOpt report $msg
        } else {
            puts $msg
        }
    }
}


proc setVerbosity {{verbosity 0}} {
    global G_runningHuygens G_verbosity 
    
    if {$verbosity == 0} {
        if $G_runningHuygens { huOpt verb -mode silent }
        set G_verbosity 0
    } else {
        if $G_runningHuygens { huOpt verb -mode noQs }
        set G_verbosity $verbosity
    }
}


proc printError { msg } {
    puts stderr $msg
}


# Print a DeBug (pdb) string
proc pdb {msg} {
    if [catch {open logPDB a} fileId] {
        puts "failed to open logPDB file"
    } else {
        puts $fileId $msg
        flush $fileId
	close $fileId
    }
}


proc getOpt {option} {
    global argv
    global argc

    if {[catch {
	set index [lsearch -exact $argv $option]
	if {$index >= 0 && $index < $argc -1} {
	    set res [lindex $argv [expr $index + 1]]
	} else {
	    set res "notFound"
	}
    } result]} {
	set res "notFound"
    }

    return $res
}



# File formats are identified based on their suffix:
# (uses Regular Expressions to make case-insesitive matching)
proc getFileFormat {name} {
    switch -regexp -- $name {
        ***:(?i).*\.i[cd]s  {set type ICS}
        ***:(?i).*\.tif.?   {set type TIFF}
        ***:(?i).*\.stk     {set type TIFF}
        ***:(?i).*\.pic     {set type BIORAD}
        ***:(?i).*\.ims     {set type IMARIS}
        ***:(?i).*\.imrs    {set type IMARIS}
        ***:(?i).*\.lsm     {set type LSM}
        default             {set type UNKNOWN}
    }
    #pdb "getFileFormat: format of $name: $type"
    return $type
}



proc checkImageFile { name } {

    if { ![file exists $name] } {
	myprint "Error: file $name not found\n"
	return 0
    }
    if { [getFileFormat $name] == "UNKNOWN" } {
	myprint "Error format of $name unknown\n"
	return 0
    }
    return 1
}


proc setResultsPath { path } {
    global G_Results_Path env
    set envList [array names env]
    set newPath ""
    
    if { ![file exists $path] } {
	puts "Specified results directory $path has been created \n"
       	file mkdir $path
    }    
    
    set len [string length $path]
    if { $len == 0 } {
	if {[lsearch $envList PWD] >= 0} {
	    set G_Results_Path "$env(PWD)/"
	    return
	} else {
	    myprint "Failed to find Results directory path\n" 0
	    set G_Results_Path ""
	    return
	}
    }

    # See if the first character is a '~'; if so replace it with
    # environment variable HOME. We need to do this to prevent problems
    # downstream.
    if { [string index $path 0] == "~"} {
	set trimmed [string trimleft $path ~]
	if {[lsearch $envList HOME] >= 0} {
	    set newPath "$env(HOME)$trimmed"
	} else {
	    # We can't proceed:
	    error "Failed to find home directory\n"
	}
    } else {
	set newPath $path
    }

    # See whether the last char is a '/'
    set len [string length $newPath]
    set inx [expr $len - 1]
    if { [string index $newPath $inx] != "/"} { append newPath / }

    set G_Results_Path $newPath

    myprint "Setting results directory to $G_Results_Path\n" 1
}


proc makeResultDir { } {
    global G_Results_Path
    # Create the output directory, works also if it already exists
    set dir $G_Results_Path
#for SIDB, do not create sub dir. It must exist already, usually user's tosidb
    #append dir "Results/"
    #pdb "makeResultDir: Creating $dir if necessary"
    #file mkdir $dir

    return $dir
}


# getRootName { } strips the leading path and trailing suffix from a
# fully extended filename
proc getRootName { {path noName} } {
    set length [string length $path]
    set lastSlash [string last / $path]
    if {$lastSlash == -1} {
        set start 0
    } else {
        if {$length - 1 == $lastSlash} {
            printError "Warning: $path contains no filename, substituting \"noName\""
            return noName
        }
        set start [incr lastSlash  1]
    }

    set lastDot [string last . $path]
    if {$lastDot == -1} {
        set stop [expr $length -1]
    } else {
        set stop [expr $lastDot -1]
    }

    set result [string range $path $start $stop]

    return $result
}

proc getOsName {} {
    global tcl_platform

    set os unknown
    if {[catch {
	if {[string match -nocase {IRIX*} $tcl_platform(os)]} {
	    set os Irix
	}
	if {[string match -nocase {Linux*} $tcl_platform(os)]} {
	    set os Linux
	}
    } result]} {
	puts stderr "Unable to determine which platform I'm running on."
    }
    return $os
}

proc getCpuCntIrix {} {
    set cpus 0
    if {[catch {
	set pipe [open "|hinv" r]
	set freq 0
	set unit false
	while {[gets $pipe line] >= 0} {
	    if {[string match {*MHZ*IP*Processor*} $line]} {
		if {[scan $line "%d %d MHZ %s Processor" \
			 cpus freq unit] == 3} {
		    break
		} else {
		    error
		}
	    }
	}
	close $pipe
    } result]} {
	catch {close $pipe}
    }

    return $cpus
}

proc getCpuCntLinux {} {
    set cpus 0
    if {[catch {
	set file [open "/proc/cpuinfo" r]
	while {[gets $file line] >= 0} {
	    if {[string match {processor*} $line]} {
		incr cpus
	    }
	}
	close $file
    } result]} {
	catch {close $file}
    }

    return $cpus
}

proc getCpuCnt {{max 1024}} {
    global tcl_platform

    set cpus 0
    if {[catch {
	set os [getOsName]
	switch $os {
	    Irix {
		set cpus [getCpuCntIrix]
	    }
	    Linux {
		set cpus [getCpuCntLinux]
	    }
	    default {}
	}
    } result]} {
    }
    if {$cpus < 1} {
	puts stderr \
	    "Failed to establish cpu count; setting to 2"
	set cpus 2
    }
    if {$cpus > $max} { 
	myprint "$cpus CPUs detected, limiting to $max"
	set cpus $max 
    }

    return $cpus
}

proc pipeToTaskId {pipe} {
    global G_taskSlot G_concurrentJobs 
    
    set inx 0
    set found 0
    set taskId -1
    while {$inx < $G_concurrentJobs && !$found} {
        set tmpPipe [lindex $G_taskSlot($inx) 1]
        
        if { $tmpPipe == $pipe} {
            set found 1
            set taskId [lindex $G_taskSlot($inx) 0]
        }
        incr inx 1
    }
    #pdb "pipeToTaskId: translated $pipe to task $taskId"
    return $taskId
}


proc pipeToSlotId {pipe} {
    global G_taskSlot G_concurrentJobs 
    
    set inx 0
    set found 0
    set slotId -1
    while {$inx < $G_concurrentJobs && !$found} {
        set tmpPipe [lindex $G_taskSlot($inx) 1]
        
        if { $tmpPipe == $pipe} {
            set found 1
            set slotId $inx
        } else {
            incr inx 1
        }
    }
    #pdb "pipeToSlotId: translated $pipe to slot id $slotId"
    return $inx
}


proc appendTaskInfo {taskId key msg} {
    global G_taskInfo
    
    
    if  {$G_taskInfo($taskId) == "0"} {
        set G_taskInfo($taskId) [list $key $msg]
    } else {
        lappend  G_taskInfo($taskId) $key $msg
    }
}


proc getValueFromList {mylist key} {
    set cnt [llength $mylist]
    set inx [lsearch -exact $mylist $key] 
    
    if {$inx < 0} {
        return 0
    }
    
    incr inx 1
    if {$inx >= $cnt} {
        printError "getValueFromList: failed to find value string to key $key in
        $mylist, returning 0"
        set res 0
    } else {
        set res [lindex $mylist $inx]
    }
    if {[string first - $res] == 0} {
	printError "getValueFromList: value for key $key starts with '-': probably is a key"
    }
    return $res
}


proc getTaskInfo {taskId key} {
    global G_taskInfo
    
    set str  $G_taskInfo($taskId)
    set res [getValueFromList $str $key]
    
    return $res
}


# Delete an option-value pair from mylist and return the reduced list
proc deleteOptionValue {mylist key} {
    
    set cnt [llength $mylist]
    set inx [lsearch -exact $mylist $key] 
    if {$inx < 0} {
        printError "deleteOptionValue: key $key not found in info for task $taskId"
        return $mylist
    }
    set valInx [expr $inx + 1]
    if { $valInx >= $cnt } {
        printError "deleteOptionValue: nothing found after key $key in info for task $taskId"
        return $mylist
    }
    set val [lindex $mylist $valInx]
    if {[string first - $val] == 0} {
	printError "deleteTaskInfoOptionValue: value for key $key starts with '-': probably is a key"
	incr valInx -1
    }
    
    # Delete the key and value
    set newList [lreplace $mylist $inx $valInx]
    # pdb "deleteOptionValue: new task info string: $newList"
    
    return $newList
}


# When an information class can be multiple present (like the '-info' class),
# this function will retrieve all
proc getTaskInfoAll {taskId key} {
    global G_taskInfo
    
    set mylist $G_taskInfo($taskId)
    
    set done 0
    set cnt 0
    set valueList ""
    
    while {!$done && $cnt < 100} {
        set val [getValueFromList $mylist $key]
        if {$val != "0"} {
            lappend valueList $val
            set mylist [deleteOptionValue $mylist $key]
        } else {
            set done 1
        }
        incr cnt 1
    }
    
    return $valueList
}

# Expand a single value or short list to a 4 element vector:
proc expandListToVec { pList varName} {
    set cnt [llength $pList]
    if { $cnt > 4} { set cnt 4}
    
    for {set i 0} { $i < $cnt } {incr i } {
	lappend outList [lindex $pList $i]
    }
    # Append the last element until we have 4 elements
    for {set i $cnt} { $i < 4 } {incr i } {
	lappend outList [lindex $pList end]
    }
    #pdb "expandListToVec: $outList for $varName"
    return $outList
}

# ------------------------ End of small utilities --------------------------	



proc findScriptDir { scriptName } {
    global env
    
    if { [file exists $scriptName]} {
        return "[pwd]/"
    } elseif { [llength [array names env SVISCRIPTS]] } {
        set scriptDir $env(SVISCRIPTS)/
    } elseif { [llength [array names env SVIHOME]] } { 
        set scriptDir $env(SVIHOME)/Scripts/
    } else {
        set scriptDir /usr/local/svi/Scripts/
    }
    return $scriptDir
}


proc initComputeEnvironment { } {
    global G_concurrentJobs computeEnvironment env
    
    set cpus [getCpuCnt]

    if { $computeEnvironment(ConcurrentJobs) == "auto" && \
	     $computeEnvironment(ThreadsPerJob) == "auto" } {
	set G_concurrentJobs $cpus
	set threads 1
    } elseif { $computeEnvironment(ConcurrentJobs) == "auto" } {
	set threads $computeEnvironment(ThreadsPerJob)
	set G_concurrentJobs [expr $cpus / $threads]
    } elseif { $computeEnvironment(ThreadsPerJob) == "auto" } {
	set G_concurrentJobs $computeEnvironment(ConcurrentJobs)
	set threads [expr $cpus/$G_concurrentJobs]
    } else {
	set G_concurrentJobs $computeEnvironment(ConcurrentJobs)
	set threads $computeEnvironment(ThreadsPerJob)
    }
    
    
    # Sanitize the result from the above or all-manual settings:

    if { $G_concurrentJobs > $cpus } {
	set G_concurrentJobs $cpus 
    }
    if { $G_concurrentJobs < 1 } {
	set G_concurrentJobs 1
    }

    if { $threads > $cpus } {
	set threads $cpus
    }
    if { $threads < 1 } {
	set threads 1
    }

    # When running a lot of jobs it is probably best to set OMP_DYNAMIC on so
    # the system can adjust the load if necessary:
    set env(OMP_DYNAMIC) TRUE
    set env(OMP_NUM_THREADS) $threads

    set totalTreads [expr $G_concurrentJobs * $threads]
    myprint "Aiming to run max $G_concurrentJobs concurrent jobs each on $threads CPUs"
    
}


proc estimateOptimalThreadCnt { remainingTasks } {
    set cpus [getCpuCnt]
    set tuning 0.5
    
    # Setting the thread count to $cpus/$remainingTasks is not smart since
    # at the end of the run it is likely tasks will then remain which are
    # started on 1 cpu.
    # Instead, use a Gaussian with sigma 'cpus'
    set garg [expr ($remainingTasks - 1.0) * ($remainingTasks - 1.0)]
    set denom [expr $cpus * $cpus * 2.0 * $tuning]
    set threads [expr round(1.0 + ($cpus - 1.0) * exp(- $garg /$denom) ) ]
    if { $threads > $cpus } {
	set threads $cpus
    }
    if { $threads < 1 } {
	set threads 1
    }
    #pdb "estimateOptimalThreadCnt: $threads"
    
    return $threads
}


# When there are less tasks left than the number of desired concurrent
# jobs the thread count can be increased.
# Do this only in 'auto' mode
#
proc updateMaxTreadCnt { remainingTasks } {
    global G_concurrentJobs computeEnvironment env
    
    if {$computeEnvironment(ThreadsPerJob) != "auto" } {
	return
    }
    set cpus [getCpuCnt]
    if {$remainingTasks >= $G_concurrentJobs + $cpus} {
	# No reason to increase thread count
        return
    }

    set oldThreads $env(OMP_NUM_THREADS)

    set threads [estimateOptimalThreadCnt $remainingTasks]

    if {$oldThreads != $threads} {
	myprint "Adapting thread count for remaining $remainingTasks jobs\
            from $oldThreads to $threads"
        set env(OMP_NUM_THREADS) $threads
    }
}

proc getHuBin { } {
    global tcl_platform
    if {$tcl_platform(os) == "Darwin"} {
	return {/Applications/SVI/bin/huscript.remote -noTk}
    } elseif  {$tcl_platform(platform) == "windows"} {
	return {"C:/Program\ Files/SVI/bin/huscript.exe" -noTk}
    } else {
	return {/usr/local/bin/huscript -noTk}
    }
}

proc huBatch_init { } {
    global G_taskCnt G_huscriptVersion G_ScriptDir G_SubTaskScriptPath \
	G_concurrentJobs 
    
    set G_taskCnt 0
    set G_huscriptVersion [getHuBin]

    isHuygens
    
    
    # Set the name of the sub-task script for this entire script: 
    set subTaskScript restOneImage.tcl
    set G_ScriptDir [findScriptDir $subTaskScript]
    set G_SubTaskScriptPath $G_ScriptDir
    append G_SubTaskScriptPath $subTaskScript
    
    # Worst case number of concurrent jobs
    # G_concurrentJobs will be adjusted lateron
    set G_concurrentJobs [getCpuCnt]
    for { set i 0} { $i < $G_concurrentJobs } { incr i 1 } {
	set G_taskSlot($i) "0"
    }
    
    set G_state "initialized"
}


# For all savety, set the default parameters:
# Also take care of other  initializations.
#
proc setDefaultParams {ctl} {
    # The global variables defaultMP and defaultRP have no prefix G_ because the
    # also show up in the user editable control file and might make it look
    # complicated.
    global defaultMP defaultRP G_taskCnt

    
    
    #   Set the default microscopic parameters:
    set defaultMP(NumericalAperture) 1.3
    set defaultMP(RefractiveIndexLensMedium) 1.515
    set defaultMP(RefractiveIndexSpecimenMedium) 1.515
    set defaultMP(ExcitationWavelength) 488.0
    set defaultMP(EmissionWavelength) 520.0
    set defaultMP(PhotonCount) 1
    # Unit: micrometers
    set defaultMP(LateralSampling) 0.05 		
    set defaultMP(AxialSampling) 0.2
    # Unit: nanometers, backprojected	
    set defaultMP(PinholeRadius) 0.28
    # A choice from confocal|widefield|4Pi|nipkow	
    set defaultMP(MicroscopeType) widefield 
    #Pinhole spacing of the nipkow disk 
    set defaultMP(PinholeSpacing) 2.53

    #	Set the default restoration parameters:
    set defaultRP(Method) cmle
    set defaultRP(BackgroundMode) auto
    set defaultRP(BackgroundPerc) 0.0
    set defaultRP(SignalToNoise) 30
    set defaultRP(MaxIterations) 100
    set defaultRP(QualityCriterion) 0.1
    set defaultRP(Padding) auto
    set defaultRP(Timeout) 7200
}


proc setRootDirectory { path } {
    global G_ROOT env
    set envList [array names env]
    set newPath ""

    if [file exists $path] {
        set len [string length $path]
        if { $len == 0 } {
	    if {[lsearch $envList PWD] >= 0} {
		set G_ROOT "$env(PWD)/"
		return
	    } else {
		myprint "Failed to find working directory\n" 0
		set G_ROOT ""
		return
	    }
        }

        # See if the first character is a '~'; if so replace it with
        # environment variable HOME. We need to do this to prevent problems
        # downstream.
        if { [string index $path 0] == "~"} {
	    set trimmed [string trimleft $path ~]
	    if {[lsearch $envList HOME] >= 0} {
		set newPath "$env(HOME)$trimmed"
	    } else {
		# We can't proceed:
		error "Failed to find home directory\n"
	    }
        } else {
	    set newPath $path
        }

        # See whether the last char is a '/'
        set len [string length $newPath]
        set inx [expr $len - 1]
        if { [string index $newPath $inx] != "/"} { append newPath / }

        set G_ROOT $newPath
    } else {
	error "Specified root directory $path does not exists.\n"
    }
    myprint "Setting root directory to $G_ROOT\n" 1
}

proc addRootDir { name } {
    global G_ROOT

    set fullPath $G_ROOT
    append fullPath $name
    return $fullPath
}

# registerMPSet{} and registerRPSet{} are called from the ctl file to register
# parameter sets

proc registerMPSet { myset} {
    global G_MPList
    upvar $myset arr
    global $myset

    # Copy the contents of the local-scope array to a global array:	
    #set content [array get arr]
    array set $myset [array get arr]
    
    lappend G_MPList $myset
}

proc registerRPSet { myset} {
    global G_RPList
    upvar $myset arr
    global $myset
    
    # Copy the contents of the local-scope array to a global array:	
    array set $myset [array get arr]
    lappend G_RPList $myset
}


proc addSingleTask {fullPath psf MPSet RPSet } {
    
    global G_taskName G_taskStatus G_taskPSF G_taskMPSet G_taskRPSet G_taskCnt \
        G_taskInfo
    
    myprint "Adding task $fullPath, task ID: $G_taskCnt" 

    if {![checkImageFile $fullPath]} { 
	error "Problem with adding task encountered\n" 
	return 0
    }

    if {$psf == "auto"} {
        set psfPath $psf
    } else {
        set psfPath [addRootDir $psf]
	if {![checkImageFile $psfPath]} {
	    error "Problem with adding task encountered\n"
            return 0
        }
    }
    
    set G_taskName($G_taskCnt) $fullPath
    set G_taskPSF($G_taskCnt) $psfPath
    set G_taskMPSet($G_taskCnt) $MPSet
    set G_taskRPSet($G_taskCnt) $RPSet
    set G_taskStatus($G_taskCnt) "queued"
    set G_taskInfo($G_taskCnt) "0"
    incr G_taskCnt 1

    return 1
}


proc setExcludePattern { pattern } {
    global G_ExcludePattern
    set G_ExcludePattern $pattern
}


proc filterPathList { pathList } { 
    global G_ExcludePattern
    if { $G_ExcludePattern == "none" } { 
	return $pathList
    }
    set outList ""
   
    foreach item $pathList {
	if { [string match [addRootDir $G_ExcludePattern] $item]} {
	    myprint "filterPathList: excluding $item\n"
	} else {
	    lappend outList $item
	}
    }
    return $outList
}


proc expandPathPattern {inPathPattern} {

    set fullPattern [addRootDir $inPathPattern]
    set pathList [glob $fullPattern]
    set pathList [filterPathList $pathList]
    
    # getFileFormat looks at the suffix only so we can use it on the pattern:
    if {[getFileFormat $inPathPattern] == "TIFF"} {
        # If there are wildcard characters in the pattern it might
        # cause problems because of conflicts with the automatic
        # reading of numbered TIFF series.
        if [regexp {[][ \* \?]+} $inPathPattern] {
            # For now, just generate a warning:
            myprint "Warning: pattern $inPathPattern will be expanded to:\n"
	    foreach item $pathList {
		myprint "$item\n"
	    }
            # myprint "Is this OK? (type y or n)\n"
            # gets stdin answer
            #if {$answer == "n"} {
            #    return ""
            #}
        }
    }
    
    return $pathList
}


# addTask{} is called from the ctl file to add tasks to the queue
#
proc addTask {inPathPattern {psf auto} {MPSet -} {RPSet -} } {
    set pathList [expandPathPattern $inPathPattern]
    if {[llength $pathList] == 0} {
        myprint "No image files found: skipping task"
    }
    
    foreach  item $pathList {
        addSingleTask $item $psf $MPSet $RPSet
    }
}

proc constructMPStrFromArr {arr itemList} {
    global defaultMP
    global $arr
    set str ""

    set trueItems [array names $arr]
    
    foreach item $itemList {
	set inx [lsearch $trueItems $item]
	if { $inx < 0 } {
	    error "Unknown microscopic parameter $item."
	    return
	}
    }

    if {[lsearch $itemList NumericalAperture] >= 0} { 
	set tmp [array get $arr NumericalAperture]
	append str "-na [lindex $tmp 1] "
    }
    if {[lsearch $itemList RefractiveIndexLensMedium] >= 0} { 
	set tmp [array get $arr RefractiveIndexLensMedium]
	append str "-ril [lindex $tmp 1] "
    }
    if {[lsearch $itemList RefractiveIndexSpecimenMedium] >= 0} { 
	set tmp [array get $arr RefractiveIndexSpecimenMedium]
	append str "-ri [lindex $tmp 1] "
    }
    if {[lsearch $itemList ExcitationWavelength] >= 0} { 
	set tmp [array get $arr ExcitationWavelength]
	append str "-ex \{ [lindex $tmp 1] \} "
    }
    if {[lsearch $itemList EmissionWavelength] >= 0} { 
	set tmp [array get $arr EmissionWavelength]
	append str "-em \{ [lindex $tmp 1]  \} "
    }
    if {[lsearch $itemList PhotonCount] >= 0} { 
	set tmp [array get $arr PhotonCount]
	append str "-pcnt \{ [lindex $tmp 1]  \} "
    }
    if {[lsearch $itemList PinholeRadius] >= 0} { 
	set tmp [array get $arr PinholeRadius]
	append str "-pr \{ [lindex $tmp 1]  \} "
    }
    if {[lsearch $itemList MicroscopeType] >= 0} { 
	set tmp [array get $arr MicroscopeType]
	append str "-micr [lindex $tmp 1] "

        if [string match [lindex $tmp 1] "nipkow"] {
	    if {[lsearch $itemList PinholeSpacing] >= 0} {
		set tmp [array get $arr PinholeSpacing]
		append str "-ps [lindex $tmp 1] "
	    }
        } else {
	    append str "-ps 1.000 " 
	}
    }
    # Only add the sampling string if one or both sampling parameters are
    # specified. Check for their specification first:
    if {[lsearch $itemList LateralSampling] >= 0} { 
	set tmp [array get $arr LateralSampling]
	set dLat [lindex $tmp 1]
    } else {
	set dLat -1
    }

    if {[lsearch $itemList AxialSampling] >= 0} { 
	set tmp [array get $arr AxialSampling]
	set dAx [lindex $tmp 1]
    } else { 
	set dAx -1
    }

    if {$dLat > 0.0 || $dAx > 0.0} {
	if {$dLat <= 0.0} {
	    set dLat $defaultMP(LateralSampling)
	}
	if {$dAx <= 0.0} {
	    set dAx $defaultMP(AxialSampling)
	}
	append str "-s {$dLat $dLat $dAx } "
    }
    
    #pdb "constructMPStrFromArr: constructed:$str"

    return $str
}


proc createUnknown_MPStr {inx} {
    global defaultMP
    
    set itemList "NumericalAperture "
    append itemList "RefractiveIndexLensMedium RefractiveIndexSpecimenMedium "
    append itemList "ExcitationWavelength EmissionWavelength PhotonCount "
    append itemList "LateralSampling AxialSampling "
    append itemList "PinholeRadius MicroscopeType "
    append itemList "PinholeSpacing"

    return [constructMPStrFromArr defaultMP $itemList]
}

proc createIMS_MPStr {inx} { 

    set itemList "NumericalAperture "
    append itemList "RefractiveIndexLensMedium RefractiveIndexSpecimenMedium "
    append itemList "ExcitationWavelength EmissionWavelength PhotonCount "
    append itemList "PinholeRadius MicroscopeType "
    append itemList "PinholeSpacing"
    
    return [constructMPStrFromArr defaultMP $itemList]
}

proc createBiorad_MPStr {inx} { 
    
    set itemList "NumericalAperture "
    append itemList "RefractiveIndexLensMedium RefractiveIndexSpecimenMedium "
    append itemList "ExcitationWavelength EmissionWavelength PhotonCount "
    append itemList "PinholeRadius "
    
    set str [constructMPStrFromArr defaultMP $itemList]
    
    # Biorads are always confocal:
    append str " -micr confocal "
    
    return $str
}

proc createTIFF_MPStr {inx} {
    return [createUnknown_MPStr $inx]
}

proc createLSM_MPStr {inx} {
    return [createUnknown_MPStr $inx]
}      


proc getAuxMPString {inx} {
    # First get hold of the name of the array holding the set:
    global G_taskMPSet
    # Get the name of the global array holding the micr. parameters
    global $G_taskMPSet($inx)
    set arr $G_taskMPSet($inx)
    if { $arr == "-"} {
	return "nil"
    }
    
    set itemList [array names $arr]
    if {$itemList == ""} {error "Specified MP set is empty: check spelling!"}
    #pdb "getAuxMPString: items from array $arr in list: $itemList\n"
    set str [constructMPStrFromArr $arr $itemList]
    
    return $str
}


proc createMPString {inx} {
    global G_taskName defaultMP
    set fileType [getFileFormat $G_taskName($inx)]
    set optionStr " "

    # Depending on the file type we substitute a number of
    # microscopic parameters from the default settings:
    switch -exact $fileType {
	ICS 	{append  optionStr "nil"}
	TIFF	{append  optionStr [createTIFF_MPStr $inx]}
	IMARIS 	{append  optionStr [createIMS_MPStr $inx]}
	BIORAD	{append  optionStr [createBiorad_MPStr $inx]}
	LSM	{append  optionStr [createLSM_MPStr $inx]}
	default {append  optionStr [createUnknown_MPStr $inx]}
    }
    
    return $optionStr
}


proc createRPString {inx method} {
    # First get hold of the name of the array holding the set:
    global G_taskRPSet defaultRP
    # Get the name of the global array holding the micr. parameters
    global $G_taskRPSet($inx)
    set arr $G_taskRPSet($inx)
    if {$arr == "-"} {
	return ""
    }

    set itemList [array names $arr]
    if {$itemList == ""} {error "Specified RP set is empty: check spelling!"}
    set trueItems [array names defaultRP]
    #pdb "getAuxRPString: items from array $arr in list: $itemList\n"
    foreach item $itemList {
	set linx [lsearch $trueItems $item]
	if { $linx < 0 } {
            if {$item == "Background"} {
                myprint "Warning: found obsolete \"Background\" field.\
                    See comments in batchCtl.tcl"
            } else {
		error "Unknown restoration parameter $item."
            }
	}
    }

    set str " "

    if {[lsearch $itemList SignalToNoise] >= 0} { 
	set tmp [lindex [array get $arr SignalToNoise] 1]
    } else {
	set tmp $defaultRP(SignalToNoise)
    }
    
    if {$method == "qmle"} {
        if {$tmp == "low" || $tmp == "fair" || $tmp == "good" \
		|| $tmp == "inf"} {
            append str "-sn $tmp "
        } else {
            if {$tmp < 20} {
                set snrStr "low"
            } elseif {$tmp < 30} {
                set snrStr "fair"
            } elseif {$tmp < 70} {
                set snrStr "good"
            } else {
                set snrStr "inf"
            } 
            append str "-sn $snrStr "
            myprint "Warning: bad SNR setting for method qmle: $tmp, should\nbe\
                low, fair, good or inf. Selecting \"$snrStr\" for now."
        }
    } else {
	set tmp [expandListToVec $tmp SignalToNoise]

        append str "-sn \{$tmp\} "
    }
    
    if {$method == "cmle"} {
        if {[lsearch $itemList MaxIterations] >= 0} {
            set tmp [lindex [array get $arr MaxIterations] 1]
        } else {
            set tmp $defaultRP(MaxIterations)
        }
        append str "-it $tmp "

        if {[lsearch $itemList QualityCriterion] >= 0} {
            set tmp [lindex [array get $arr QualityCriterion] 1]
        } else {
            set tmp $defaultRP(QualityCriterion)
        }
        append str "-q $tmp "
    }
    
    if {[lsearch $itemList Padding] >= 0} { 
	set tmp [lindex [array get $arr Padding] 1]
    } else {
	set tmp $defaultRP(Padding)
    }
    append str "-padding $tmp "
    
    return $str
}


proc getRPTimeout {inx} {
    global G_taskRPSet defaultRP
    # Get the name of the global array holding the micr. parameters 
    # for this task
    global $G_taskRPSet($inx)
    set item Timeout
    set defaultTimeout $defaultRP(Timeout)
    
    set arr $G_taskRPSet($inx)
    if {$arr == "-"} {
	return $defaultTimeout
    }
    set trueItems [array names defaultRP]
    set itemList [array names $arr]
    
    set linx [lsearch $trueItems $item]
    if { $linx < 0 } {
        error "getRPTimeout: Unknown restoration parameter $item."
    }
    if {[lsearch $itemList $item] >= 0} { 
	set tmp [lindex [array get $arr $item] 1]
    } else {
	set tmp $defaultRP($item)
    }
    #pdb "getRPTimeout: $tmp secs"
    return $tmp
}


proc getRestoMethod {inx} {
    # First get hold of the name of the array holding the set:
    global G_taskRPSet defaultRP
    # Get the name of the global array holding the micr. parameters
    global $G_taskRPSet($inx)
    set arr $G_taskRPSet($inx)

    if {$arr == "-"} {
	return $defaultRP(Method)
    }

    set itemList [array names $arr]	
    if {[lsearch $itemList Method] >= 0} { 
	set tmp [array get $arr Method]
	set tmp2 [lindex $tmp 1]
	if { !($tmp2 == "cmle" || $tmp2 == "qmle") } {
	    myprint "Substituting \"cmle\" for bad restoration method $tmp2"
	    set tmp2 cmle
	}	
    } else {
	set tmp2 $defaultRP(Method)
    }
    return $tmp2
}


proc checkOldStyleBg {inx } {
    global G_taskRPSet defaultRP

    # Get the name of the global array holding the micr. parameters
    global $G_taskRPSet($inx)
    set arr $G_taskRPSet($inx)
    set itemList [array names $arr]

    if {[lsearch $itemList Background] >= 0} {
	myprint "Found obsolete \"Background\" field."
    } else {
        return -1
    }
    if {[lsearch $itemList BackgroundPerc] >= 0} {
        set gotPerc 1
    } else {
        set gotPerc 0
    }
    if {[lsearch $itemList BackgroundMode] >= 0} {
        set gotMode 1
    } else {
        set gotMode 0
    }
    
    if {$gotMode} {
        myprint "Ignoring obsolete \"Background\" field: see comments in batchCtl.tcl"
        return -1
    }
    
    if {[lsearch $itemList Background] >= 0} {
        set tmp [lindex [array get $arr Background] 1]
    } else {
        return -1
    }

    if { $tmp == "lowest" ||  $tmp == "object" || $tmp == "wf" || $tmp == "auto"
	 || $tmp == "manual"} {
        set background $tmp
    } else {
        if {$gotPerc} {
            
            # We have a valid setting of the background percentage; best ignore
            # the absolute setting

            myprint "Absolute value in obsolete \"Background\" value conflicts \
            with set \"BackgroundPerc\" \nvalue; using the latter .\n"
            set background $tmp
        }   
	
        if { [scan $tmp "%f" background] == 0 } {
            myprint "Strange background $tmp found: substituting 0.0\n"
            set background 0.0
        } else {
            set background $tmp
        }
        set $G_taskRPSet(BackgroundPerc) $background
        set background "manual"
    }
    
    return $background
}

proc getBackgroundMode {inx } {
    global G_taskRPSet defaultRP

    # Get the name of the global array holding the resto. parameters
    global $G_taskRPSet($inx)
    set arr $G_taskRPSet($inx)
    set itemList [array names $arr]

    if {[lsearch $itemList BackgroundMode] >= 0} { 
	set tmp [lindex [array get $arr BackgroundMode] 1]
    } else {
        # For backwards compatibility also check the older 'Background' keyword:
        set tmp [checkOldStyleBg $inx]
        if {$tmp < 0} {
            set tmp $defaultRP(BackgroundMode)
        }
    }    

    if { $tmp == "lowest" ||  $tmp == "object" || $tmp == "wf" || $tmp == "auto"
	 || $tmp == "manual"} {
	set background $tmp
    } else {
        myprint "Substituting \"lowest\" for bad background mode \"$tmp\""
        set background "lowest"
    }
    
    return $background
}


proc getBackgroundPercList {inx } {
    global G_taskRPSet defaultRP

    # Get the name of the global array holding the resto. parameters
    global $G_taskRPSet($inx)
    set arr $G_taskRPSet($inx)
    set itemList [array names $arr]

    if {[lsearch $itemList BackgroundPerc] >= 0} { 
	set val [lindex [array get $arr BackgroundPerc] 1]
	#pdb "getBackgroundPercList: $val"
    } else {
        set val $defaultRP(BackgroundPerc)
    }    
    set val [expandListToVec $val BackgroundPerc]
    return "\{ $val \}"
}


proc signalTaskEnd {taskId slotId} {
    global G_done G_taskName G_taskStatus G_taskSlot
    
    #pdb "Processing of image $G_taskName($taskId), ID $taskId, slot $slotId done,\
	# status: $G_taskStatus($taskId)"
    
    set G_taskSlot($slotId) 0
    set G_done 1
    
}


proc pipeHandler { pipe } {
    global G_done G_taskName G_state G_taskStatus G_taskSlot 
    
    
    # The eof might come at a time the process is already deleted
    # from the task slot, so its ID can no longer be determined.
    # Therefore eof handling must come first. 
    if [eof $pipe] {
        catch {close $pipe}
        pdb "pipeHandler: eof caught on pipe $pipe"
        
        #signalTaskEnd $taskId $slotId
        # return
    }
    
    set slotId [pipeToSlotId $pipe]
    if {$slotId < 0} {
        gets $pipe  trash
        printError "pipeHandler 1: output from unregistered pipe $pipe:\n\
        $trash: no action"
        return
    }
    set taskId [pipeToTaskId $pipe]
    if {$taskId < 0} {
        gets $pipe  trash
        printError "pipeHandler 2: output from unregistered pipe $pipe:\n\
        $trash: no action"
        return
    }
    
    gets $pipe line
    pdb "pipeHandler: for pipe $pipe task $taskId, slot: $slotId got: $line"
    myprint "Task $taskId reports: $line" 1
    
    switch -exact $line {
        Running     { set G_taskStatus($taskId) "Running" }
        Completed   { set G_taskStatus($taskId) "Completed"
	    #pdb "pipeHandler:Completed: task: $taskId, slot: $slotId"
	    signalTaskEnd $taskId $slotId 
	    catch {close $pipe}
	}
        Skip        { set G_taskStatus($taskId) "Aborted"
	    pdb "pipeHandler:Skip: task: $taskId, slot: $slotId"
	    signalTaskEnd $taskId $slotId 
	    catch {close $pipe}
	}
        Error       { set G_taskStatus($taskId) "Aborted"
	    signalTaskEnd $taskId $slotId
	    catch {close $pipe}
	}
    }
    
    if [string match "Output ?*" $line] {
        
        set res [lindex $line 1]
        appendTaskInfo $taskId -outfile $res
    } elseif [string match "Info ?*" $line] {
        set info [lindex $line 1]
        appendTaskInfo $taskId -info $info
    }
    
    if {$line == "Error"} {
        # A serious problem occured. Stop starting new processes by setting
        # global state to 'fatal error'.
        set G_state "Fatal error"
    }
    
    return
}


proc launchTask { inx resultDir } {
    global G_huscriptVersion
    global G_SubTaskScriptPath
    global G_taskName G_taskPSF G_done G_state G_taskStatus
    global env
    set failedToStart 0
    
    if {$G_state == "Fatal error"} {
        return -1
    }

    # Because all recognizable files on the command line are read AND removed
    # we add a trailing 'X' here. Function stripDummyExtention in restOneImage.tcl
    # strips it:
    
    set orgStr "-org $G_taskName($inx)X"
    set tmp     [createMPString $inx]

    set mp1Cmd  "-mp1Cmd  {$tmp}"
    set mp2Cmd  "-mp2Cmd  {[getAuxMPString $inx]}"

    set psfOpt  "-psfPath $G_taskPSF($inx)X"
    set method  [getRestoMethod $inx]
    set methodStr  "-method $method"
    set rpOpt   "-rpOpt {[createRPString $inx $method]}"
    set bgModeOpt   "-bgMode [getBackgroundMode $inx]"
    set bgPercOpt   "-bgPerc [getBackgroundPercList $inx]"
    set taskId  "-taskID $inx"
    set ppid    "-ppid [pid]"
    # It seems that the OMP env. variables are not inherited so pass them on:
    set ompOpt  "-ompDyn $env(OMP_DYNAMIC) -ompThreads $env(OMP_NUM_THREADS)"
    set resultDir "-resDir $resultDir"
    set myargs "-script $G_SubTaskScriptPath \
        $orgStr $mp1Cmd $mp2Cmd $psfOpt \
        $rpOpt $bgModeOpt $bgPercOpt $taskId $methodStr \
        $ppid $ompOpt $resultDir"

    set pipe [open "|$G_huscriptVersion $myargs" r]
    # It is possible the launch is successful, but an error occurs immediately
    # for instance because of a missing license. 
    # To catch this read the first line and check on an eof:
    if { ![eof $pipe] } {
	gets $pipe xxDummy
	if [eof $pipe] {
	    set failedToStart 1
	}
    } else {
	set failedToStart 1
    }
    
    if {$failedToStart} {
	set G_taskStatus($inx) "Failed to start"
	pdb "Failed to start $G_huscriptVersion.\nTips:\tCheck by\
                running $G_huscriptVersion from a shell.\n\tCheck license"
	if {[string length $env(SVIHOME)] > 0} {
	    pdb "\tThe SVIHOME environment variable is set\
                         to \"$env(SVIHOME)\".\
                         \n\tCheck whether this is correct"
	}
	return -1
    }
    #pdb "Task $inx gets pipe: $pipe"
    fconfigure $pipe -blocking 0
    fileevent $pipe readable [list pipeHandler $pipe]
    
    set G_taskStatus($inx) "Launched"
    
    #pdb "Task $G_huscriptVersion $myargs with task ID $inx launched"
    return "$pipe"
}


proc killRunningTasks { reason } {
    global G_taskSlot G_concurrentJobs G_taskName G_taskStatus
    
    
    for { set i 0} { $i < $G_concurrentJobs } { incr i 1 } {
        #pdb "killRunningTasks: slot $i: $G_taskSlot($i)"
        if { $G_taskSlot($i) != "0" } {
            set childPipe [lindex $G_taskSlot($i) 1]
            set childPid [pid $childPipe]
            set taskId [lindex $G_taskSlot($i) 0]
            
            set cmd "exec kill -9 $childPid"
            myprint "killRunningTasks: executing command: $cmd..." 1
            eval $cmd
            myprint "Task $G_taskName($taskId), ID $taskId, killed due to timeout" 1
	    set G_taskSlot($i) "0"
            set G_taskStatus($taskId) $reason
            
            catch {close $childPipe} pipeCloseResult
            #pdb "killRunningTasks pipe close: $pipeCloseResult"
        }
    }
}


proc timeOut { msec } {
    global G_timeout G_done
    
    set sec [expr $msec / 1000.0]
    set msg [format "timeOut handler: timeout after %.1f seconds" $sec]
    set G_timeout 1 
    set G_done 1
    pdb "$msg"
}


proc getIdleTask { } {
    global G_taskSlot G_concurrentJobs
    
    set inx 0
    
    # See whether there is an empty slot:
    while {$inx < $G_concurrentJobs && $G_taskSlot($inx) != "0"} {
        incr inx 1
    }
    
    if {$inx == $G_concurrentJobs} {
        return -1
    }
    
    return $inx
}


proc getActiveTasks {} {
    global G_taskSlot G_concurrentJobs
    set cnt 0

    for {set i 0} {$i < $G_concurrentJobs} {incr i 1} {
        if {$G_taskSlot($i) != "0"} {
            incr cnt 1
        }
    }
    
    return $cnt
}


proc getMaxActiveTaskTimeout {} {
    global G_concurrentJobs G_taskSlot
    set timeoutSeconds 0
    
    for {set i 0} {$i < $G_concurrentJobs} {incr i 1} {
        if {$G_taskSlot($i) != "0"} {
            set taskId [lindex $G_taskSlot($i) 0]
            
            # Compute the largest timeout of active tasks
            set taskTimeOutSecs [getRPTimeout $taskId] 
            if { $taskTimeOutSecs > $timeoutSeconds } {
                set timeoutSeconds $taskTimeOutSecs
            }
        }
    }
    #pdb "getMaxActiveTaskTimeout: $timeoutSeconds"
    return $timeoutSeconds
}


proc enterEventLoop  {resultDir } {
    global G_taskCnt G_concurrentJobs G_taskSlot G_taskStatus G_taskName \
        G_timeout G_done G_state
    
    #set activeTasks [expr $nextTaskId -1]
    set activeTasks 0
    set nextTaskId 0
    
    while  {$nextTaskId < $G_taskCnt} {
        set remainingTasks [expr $G_taskCnt - $nextTaskId]
        updateMaxTreadCnt $remainingTasks

        while {$activeTasks < $G_concurrentJobs && $nextTaskId < $G_taskCnt \
		   && $G_state != "Fatal error"} {
            set slotInx [getIdleTask]
            if {$slotInx >= 0} {
                set taskName $G_taskName($nextTaskId)
                myprint "Launching task $taskName, \
				         ID: $nextTaskId on cpu group $slotInx"
		
                # Record both task index (=task id) and child pipe in 
                # the task slot:
                set pipe [launchTask  $nextTaskId $resultDir]
                
                if {$pipe == -1} {
                    myprint "Fatal error for task $taskName, trying other tasks"
                    
                    set G_taskSlot($slotInx) 0
                    set G_taskStatus($nextTaskId) "error"
		    incr nextTaskId 1
                    continue
                }
		
		set G_taskSlot($slotInx) "$nextTaskId $pipe"

                incr activeTasks 1
                incr nextTaskId 1
                
            } else {
                error "Can't find free slot: coding error"
            }
            
        }
	
        # First set a watchdog timer:
        set timeoutSeconds [getMaxActiveTaskTimeout]
        
        set milliSecs [expr $timeoutSeconds * 1000]
        set G_done 0
        set G_timeout 0
        #pdb "enterEventLoop: setting timeout after $milliSecs ms"
        set watchDog [after $milliSecs [list timeOut $milliSecs] ]
	
        vwait G_done
        after cancel $watchDog  
        
        # pdb "enterEventLoop: vwait 1 done, timeout: $G_timeout"
        if $G_timeout {
            killRunningTasks "timeout:killed"
        }
        set activeTasks [getActiveTasks]     
    }
    
    # Now we have to wait for all the task slots to become empty.  For
    # robustness count the active processes in the list instead of
    # local bookkeeping
    set activeTasks [getActiveTasks]
    while {$activeTasks > 0} {
        set watchDog [after $milliSecs [list timeOut $milliSecs] ]
        vwait G_done
        after cancel $watchDog  
        set activeTasks [getActiveTasks]
    }
}


proc executeTasksMP {} {
    global G_taskName  G_taskCnt G_concurrentJobs G_taskSlot G_state
    
    set resultDir [makeResultDir]

    for { set i 0} { $i < $G_concurrentJobs } { incr i 1 } {
	set G_taskSlot($i) "0"
    }
    
    set resultDir [makeResultDir]

    set G_state running
    
    # Now enter the event loop. We come back here after all processing
    # has stopped.
    enterEventLoop  $resultDir 
    
    set G_state Finished    
}


proc cleanup { } {
    global G_taskMPSet G_taskRPSet G_taskCnt G_taskSlot G_taskStatus \
        G_done G_timeout

    # Go into the event loop to process remaining events, if any
    set G_timeout 0
    after 100 [list set G_timeout 1]
    vwait G_timeout
    
    unset G_taskMPSet G_taskRPSet G_taskCnt G_taskSlot G_taskStatus \
        G_done G_timeout

    file delete "restOneImage.tcl"
    file delete "doBatch.tcl"
    catch {file copy "batchCtl.tcl" "lastRun.txt"}
    file delete "batchCtl.tcl"
}


proc reportFinalState { } {
    global G_taskName  G_taskCnt G_state G_taskStatus
    
    myprint "\n\t******************* Job completion status ********************"
    for { set i 0} { $i < $G_taskCnt } { incr i 1 } {
        set storedPath [getTaskInfo $i -outfile]
        set msg "\nJob $i:\n"
        append msg "Source image:      $G_taskName($i)\n"
        append msg "Status:            $G_taskStatus($i)\n"
        set infoList [getTaskInfoAll $i -info]
        set infoCnt [llength $infoList]
        for {set j 0} {$j < $infoCnt} {incr j 1} {
            append msg "Report $j:         [lindex $infoList $j]\n"
        }
        # If the path could not be found $storedPath is 0.
        if {$storedPath != "0"} {
            append msg "Result stored in:  $storedPath"
        } else {
            append msg "No result stored."
        }
        myprint $msg
    }
    myprint "\nFinal state upon completion of all jobs: $G_state"
}


proc doJob { {ctl batchCtl.tcl} {verbosity 0} }  {
    # To prevent chicken and eggs problems set G_verbosity here
    global G_verbosity
    set G_verbosity 0
    
    # See if we need to print online help:
    set helpStr "syntax: doJob <controlFile>"
    if { $ctl == "?"} {
	huOpt report $helpStr
	return
    }

    huBatch_init
    
    if {![checkHuScriptVersion 2 4]} {
        exit 1
    }
    
    setVerbosity $verbosity
    
    # Load the control file. Check whether it exists first:
    if { ![file exists $ctl] } {
	myprint $helpStr
	return "File $ctl not found"
    }
    
    setDefaultParams $ctl

    # Globalize the variables referenced in the control file:
    global defaultMP defaultRP computeEnvironment


    if [catch {source $ctl} result] {
	myprint "\nError in executing $ctl: $result"
	return $result
    } else {
	myprint "Control file $ctl succesfully read\n"
    }

    initComputeEnvironment

    executeTasksMP

    reportFinalState
    
    cleanup
    
    return
}


proc startDoBatch {} {
    global argv

    puts "doBatch V1.02 starting..."
    if {[catch {
	set arg1 [lindex $argv 0]
	if {[string length $arg1] == 0} {
	    set arg1 batchCtl.tcl
	}
    } result]} {
	set arg1 batchCtl.tcl
    }
    doJob $arg1
}

startDoBatch

