#   FILE restOneImage.tcl --
# 
#   Restore a single image
#   Intended to be used by supervisor script like doBatch.tcl.
#
#  
#   Author and copyright: 
#   T. Dijkhuis/H.v.d.Voort Scientific Volume Imaging Holding BV
#   Alexanderlaan 14,
#   1213 XS Hilversum,
#   The Netherlands,
#   email: info@svi.nl


# ---------------------------- Small utilities -----------------------------




proc myprint {msg} {
    if [catch {open logTasks a} fileId] {
        puts "failed to open log file"
    } else {
        puts $fileId $msg
        flush $fileId
	close $fileId
    }
    huOpt report $msg
}


proc reportToParent {msg} {
    puts stdout $msg
    flush stdout
    
    # For debugging only:
    if [catch {open logReportToParent a} fileId] {
        puts "failed to open log file"
    } else {
        puts $fileId $msg
        flush $fileId
	close $fileId
    }
}


proc getOpt {option} {
    global argv
    global argc
    
    set index [lsearch -exact $argv $option]
    
    if {$index >= 0 && $index < $argc -1} {
    	set res [lindex $argv [expr $index + 1]]
    } else {
        set res -1
        myprint "restOneImage: getOpt: unable to find option $option"
        reportToParent Error
    }
    
    return $res
}


proc stripLeadingPath {path} {
    set inx [string last / $path]
    set len [string length $path]
    
    if {$inx == -1} {
        return $path
    } else {
        if {$inx == [expr $len -1]} {
            myprint "stripLeadingPath: $path contains no filename"
            set res "noPathFound"
        }
        incr inx 1
        set res [string range $path $inx end]
    }
    #myprint "stripLeadingPath: $res"
    return $res
}


# Because all recognizable files on the command line are read AND removed
# we add a trailing 'X' in doBatch.tcl. This function strips it:
#
proc stripDummyExtention {path} {
    set stripped1 [string trimright $path]
    return [string trimright $stripped1 "X"]
}


# ------------------------- End of small utilities ---------------------------


proc cmdRestore {original method psf dest optStr bgMode bgStr} {

  #deconvolution
  set executing "$original $method $psf -> $dest $optStr -bgMode $bgMode $bgStr"
  myprint "Executing $executing"

  if [catch {eval $executing} result] {
    error "Failed to save execute command: $executing"
  }

  ## calculate maximum
  #set paramlist [$dest stat -tclReturn]
  #array set param $paramlist
  #set maxval $param(maxValue)                              
  #if {$maxval <= 255} {
  #   myprint "Executing $dest convert -type byte"
  #   $dest convert -type byte
  #} elseif {$maxval <= 32767} {                                   
  #   # convert to integer type
  #   myprint "Executing $dest convert -type int"
  #   $dest convert -type int
  #} 

}
   
proc saveImage {inx dest dir} {
    set cnt 0

    set rootName $dest
    set rootPath "$dir$rootName"
    set fullPath "$rootPath.ics"
    
    while { [file exists $fullPath] && $cnt < 100} { 
        myprint "File $fullPath exists, constructing other name..."
        set fullPath "$rootPath.$cnt.ics"
        incr cnt 1
    }
    if { $cnt == 100 } {
        reportToParent "Error"
        myprint "Unable to construct unused filename: failed to save image \
        in directory $dir"
    }
    
    set saveCmd "$dest save $fullPath -type ics"
    if [catch {eval $saveCmd} result] {
            error "Failed to save execute command: $saveCmd \n \
    result: $result. Aborting entire run"
    }
    
    return $fullPath
}

	
proc issueSetpCmds { measured argList } {   

    #myprint "issueSetpCmds: argList: $argList"
    set argc [llength $argList]
    # First issue a setp commnd for the non-channel params present in the list
    set singleParams "-na -ril -ri -s -ps -micr"
    set singleParamsOk 0
    foreach item $singleParams {
	set inx [lsearch $argList $item]
	set inxVal [expr $inx + 1]
	if { $inx > -1 && $inxVal < $argc } {
	    lappend noChanParamList $item [lindex $argList $inxVal]
	    set singleParamsOk 1
	}
    }
    if { $singleParamsOk} { 
	set cmdStr "$measured setp $noChanParamList -chan -1"
        myprint "$cmdStr"
        if [catch {eval $cmdStr} result] {
	    reportToParent "Info {Failed to set non-channel parameters to $measured}"
            reportToParent "Skip"
            return 0
        }
    }
    
    # Now the channel parameters
    set chanCnt [$measured getdims -mode ch]
    set chanParams "-ex -em -pr -pcnt"

    for {set ch 0} {$ch < $chanCnt} {incr ch 1}  { 
	set chanParamsOk 0
	set chanParamList ""
	foreach item $chanParams {
	    set inx [lsearch $argList $item]
	    set inxVal [expr $inx + 1]
	    if { $inx > -1 && $inxVal < $argc } {
		set paramValList [lindex $argList $inxVal]
		set paramValLen [llength $paramValList]

		# Only set the parameter when it is available:
		if { $ch < $paramValLen} { 
		    set val [lindex $paramValList $ch]
		    lappend chanParamList $item $val
		    set chanParamsOk 1
		}
	    }
	}
	
	if { $chanParamsOk} { 
	    set cmdStr "$measured setp $chanParamList -chan $ch"
	    myprint "$cmdStr"
	    if [catch {eval $cmdStr} result] {
		reportToParent \
		    "Info {Failed to set parameters of chan $ch of \
                         $measured}"
		reportToParent "Skip"
		return 0
	    }
	}
    }
    return 1
}


proc restoreImage {inx resultDir ppid} {
    global argv
    
    #reportToParent "Info {argv: $argv}"
    # Set the global undo system to 'off' and verbosity to 'silent'
    huOpt gundo off
    huOpt verb -mode silent

    # Get the script options related to the input file:
    set imgFilePath [stripDummyExtention [getOpt -org]]
    set setp1Opt    [getOpt -mp1Cmd]
    set setp2Opt    [getOpt -mp2Cmd]
    set taskID      [getOpt -taskID]
    if [catch {img open $imgFilePath} measured] {
        reportToParent "Info {Failed to open (1) $imgFilePath for original got $measured}"
        reportToParent "Skip"
        return 0
    }
    if [string match "Error:*" $measured] {
        myprint stderr "Failed to open image $imgFilePath: skipping this task\n"
	reportToParent "Info {Failed to open (2) $imgFilePath for original got $measured}"
        reportToParent "Skip"
        return 0
    } else {
        myprint "\n\n>>>Image $measured created"
    }

    # If a time series was imported, and there is no time option, restore 
    # merely the first frame. 
    set isTime [expr [$measured  getdims -mode t] > 0]
    set licenseList [huOpt license]
    set hasTimeLic  [expr [lsearch -exact $licenseList time] >= 0]

    if {$isTime && !$hasTimeLic} {
	$measured getframe -> $measured -frame 1
	myprint "No time license option: restoring merely the first time frame"
    } 


    myprint "Issued setp commands:"
    
    # If MParameter list is specified...
    if {![string match "*nil*" $setp2Opt]} {
        # First set default parameters,
        if {![string match "*nil*" $setp1Opt]} {
	      issueSetpCmds $measured $setp1Opt
        }
        # then set parameters from the given list.
        issueSetpCmds $measured $setp2Opt
    }
    set actualParams [$measured setp -v -tclReturn]
    myprint "actual micr. params: $actualParams"
    
    set psfPath [stripDummyExtention [getOpt -psfPath]]
    set psfName [stripLeadingPath $psfPath]
    
    if { $psfPath == "auto" } {
	# An empty PSF is sufficient:
        set psf [img create theorPSF -type float -dim {32 32 32 0} -chan 1]
    } else {
	set psfLoadCmd "img open $psfPath"
	myprint "Loading PSF with cmd: $psfLoadCmd"
	if [catch {eval $psfLoadCmd} psf] {
	    catch {$measured del}
            myprint "Failed to load PSF $psfPath for image $measured: result: $psf"
            reportToParent "Info {Failed to load PSF $psfPath for image $measured}"
            reportToParent "Skip"
            return 0
        } else {
            myprint "PSF loaded as '$psf'"
        }
	
        if [string match "Error:*" $psf] {
	    reportToParent "Info {Failed to open PSF $psfPath: task skipped}"
	    reportToParent "Skip"
            $measured del
	    return 0
	} else {
            myprint "Opened PSF $psfPath"
        }
    }
    if [catch {eval $psf stat} res] {
        myprint "could not stat psf: $res"
    }
    set methodStr [getOpt -method]
    set destName "$measured.$methodStr"
    set genDestCmd "$measured repl $destName"
    
    if [catch {eval $genDestCmd} dest] {
         # Could not replicate: serious trouble, might as well quit...
         reportToParent "Info {Failed create destination image: task skipped}"
         reportToParent "Skip"
         $measured del
         $psf del
         return 0
    }
        
    set optStr [getOpt -rpOpt]
    set bgVal [getOpt -bgPerc]
       
    set bgMode [getOpt -bgMode]
    set bgStr "-bg \{ $bgVal \}"
    myprint "Starting restoration"
    set restoCmd "cmdRestore $measured $methodStr $psf $dest {$optStr} {$bgMode} {$bgStr}"
    
    myprint "Executing command $restoCmd\n"
    set ompThreads [getOpt -ompThreads]
    set env(OMP_NUM_THREADS) $ompThreads
    set env(OMP_DYNAMIC) [getOpt -ompDyn]
    myprint "OMP environment: OMP_DYNAMIC: $env(OMP_DYNAMIC),\
      OMP_NUM_THREADS: $ompThreads"
    huOpt cpu -max $ompThreads
       
    # Do it:
    if [catch {eval $restoCmd} result] {
        myprint "Failed to execute command:\n$restoCmd \nresult: $result\n"
        reportToParent "Info {Failed to execute restoration command $restoCmd}"
        reportToParent "Skip"
        $measured del
        $psf      del
        $dest     del
        return 0
    } else {
	myprint "restoCmd ok"
    }
    
 
    # Free up space as soon as possible:
    $measured del
    $psf      del
	
    set outFile [saveImage $inx $dest $resultDir]
    reportToParent "Output {$outFile}"
    
    $dest     del

    reportToParent "Completed" 	
    return 1
}
        
            
       
   
#
# ---------------------------- Start of the script -------------------------
# 
 

#Retrieve parent's process id:
set ppid [getOpt -ppid]

huOpt report "process [pid] alive"
reportToParent "Running\n"

set status [restoreImage [getOpt -taskID] [getOpt -resDir] $ppid]
flush stdout
flush stderr

exit $status
